"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    nodeIntersection: ()=>nodeIntersection,
    intersection: ()=>intersection,
    union: ()=>union,
    flattenType: ()=>flattenType,
    flattenNode: ()=>flattenNode,
    isLiteralNode: ()=>isLiteralNode,
    arrayOf: ()=>arrayOf
});
const _astTs = require("../parse/string/ast.js");
const _errorsTs = require("../utils/errors.js");
const _genericsTs = require("../utils/generics.js");
const _pathsTs = require("../utils/paths.js");
const _composeTs = require("./compose.js");
const _predicateTs = require("./predicate.js");
const _propsTs = require("./rules/props.js");
const nodeIntersection = (l, r, state)=>{
    state.domain = undefined;
    const lResolution = state.type.meta.scope.resolveNode(l);
    const rResolution = state.type.meta.scope.resolveNode(r);
    const result = resolutionIntersection(lResolution, rResolution, state);
    if (typeof result === "object" && !(0, _genericsTs.hasKeys)(result)) {
        return (0, _genericsTs.hasKeys)(state.disjoints) ? (0, _composeTs.anonymousDisjoint)() : state.addDisjoint("domain", (0, _genericsTs.keysOf)(lResolution), (0, _genericsTs.keysOf)(rResolution));
    }
    return result === lResolution ? l : result === rResolution ? r : result;
};
const resolutionIntersection = (0, _composeTs.composeKeyedIntersection)((domain, l, r, context)=>{
    if (l === undefined) {
        return r === undefined ? (0, _composeTs.throwUndefinedOperandsError)() : undefined;
    }
    if (r === undefined) {
        return undefined;
    }
    return (0, _predicateTs.predicateIntersection)(domain, l, r, context);
}, {
    onEmpty: "omit"
});
const intersection = (l, r, type)=>{
    const state = new _composeTs.IntersectionState(type, "&");
    const result = nodeIntersection(l, r, state);
    return (0, _composeTs.isDisjoint)(result) ? (0, _errorsTs.throwParseError)((0, _astTs.compileDisjointReasonsMessage)(state.disjoints)) : (0, _composeTs.isEquality)(result) ? l : result;
};
const union = (l, r, type)=>{
    const lResolution = type.meta.scope.resolveNode(l);
    const rResolution = type.meta.scope.resolveNode(r);
    const result = {};
    const domains = (0, _genericsTs.keysOf)({
        ...lResolution,
        ...rResolution
    });
    for (const domain of domains){
        result[domain] = (0, _genericsTs.hasKey)(lResolution, domain) ? (0, _genericsTs.hasKey)(rResolution, domain) ? (0, _predicateTs.predicateUnion)(domain, lResolution[domain], rResolution[domain], type) : lResolution[domain] : (0, _genericsTs.hasKey)(rResolution, domain) ? rResolution[domain] : (0, _composeTs.throwUndefinedOperandsError)();
    }
    return result;
};
const hasImpliedDomain = (flatPredicate)=>flatPredicate[0] && (flatPredicate[0][0] === "value" || flatPredicate[0][0] === "class");
const flattenType = (type)=>{
    const ctx = {
        type,
        path: new _pathsTs.Path(),
        lastDomain: "undefined"
    };
    return type.meta.config ? [
        [
            "config",
            {
                config: type.meta.config,
                node: flattenNode(type.node, ctx)
            }
        ]
    ] : flattenNode(type.node, ctx);
};
const flattenNode = (node, ctx)=>{
    if (typeof node === "string") {
        const resolution = ctx.type.meta.scope.resolve(node);
        const updatedScopeConfig = resolution.meta.scope !== ctx.type.meta.scope && resolution.meta.scope.config;
        return updatedScopeConfig ? [
            [
                "config",
                {
                    config: updatedScopeConfig,
                    node: resolution.flat
                }
            ]
        ] : resolution.flat;
    }
    const domains = (0, _genericsTs.keysOf)(node);
    if (domains.length === 1) {
        const domain = domains[0];
        const predicate = node[domain];
        if (predicate === true) {
            return domain;
        }
        ctx.lastDomain = domain;
        const flatPredicate = (0, _predicateTs.flattenPredicate)(predicate, ctx);
        return hasImpliedDomain(flatPredicate) ? flatPredicate : [
            [
                "domain",
                domain
            ],
            ...flatPredicate
        ];
    }
    const result = {};
    for (const domain1 of domains){
        ctx.lastDomain = domain1;
        result[domain1] = (0, _predicateTs.flattenPredicate)(node[domain1], ctx);
    }
    return [
        [
            "domains",
            result
        ]
    ];
};
const isLiteralNode = (node, domain)=>{
    return (0, _predicateTs.resolutionExtendsDomain)(node, domain) && (0, _predicateTs.isLiteralCondition)(node[domain]);
};
const arrayOf = (node)=>({
        object: {
            class: "Array",
            props: {
                [_propsTs.mappedKeys.index]: node
            }
        }
    });
