"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
Object.defineProperty(exports, "parseKeyOfTuple", {
    enumerable: true,
    get: ()=>parseKeyOfTuple
});
const _predicateTs = require("../../nodes/predicate.js");
const _propsTs = require("../../nodes/rules/props.js");
const _errorsTs = require("../../utils/errors.js");
const _genericsTs = require("../../utils/generics.js");
const _numericLiteralsTs = require("../../utils/numericLiterals.js");
const _definitionTs = require("../definition.js");
const _astTs = require("../string/ast.js");
const numericIndexKey = {
    regex: _numericLiteralsTs.wellFormedNonNegativeIntegerMatcher.source
};
const parseKeyOfTuple = (def, ctx)=>{
    const resolution = ctx.type.meta.scope.resolveNode((0, _definitionTs.parseDefinition)(def[1], ctx));
    if (!(0, _predicateTs.resolutionExtendsDomain)(resolution, "object")) {
        return (0, _errorsTs.throwParseError)((0, _astTs.writeImplicitNeverMessage)(ctx.path, "keyof"));
    }
    if (resolution.object === true) {
        return (0, _astTs.writeImplicitNeverMessage)(ctx.path, "keyof");
    }
    const objectBranches = (0, _genericsTs.listFrom)(resolution.object);
    let sharedKeys = keysOfBranch(objectBranches[0]);
    for(let i = 1; i < objectBranches.length; i++){
        const branchKeys = keysOfBranch(objectBranches[i]);
        // we can filter directly by equality here because the RegExp we're
        // using will always be reference equal to
        // wellFormedNonNegativeIntegerMatcher
        sharedKeys = sharedKeys.filter((k)=>branchKeys.includes(k));
    }
    const result = sharedKeys.map((k)=>typeof k === "string" ? {
            value: k
        } : {
            regex: k.source
        });
    return result.length ? {
        string: result.length === 1 ? result[0] : result
    } : (0, _astTs.writeImplicitNeverMessage)(ctx.path, "keyof");
};
const keysOfBranch = (branch)=>{
    if (!("props" in branch)) {
        return [];
    }
    if (branch.class !== "Array") {
        return Object.keys(branch.props);
    }
    const arrayProps = [];
    for (const key of Object.keys(branch.props)){
        if (key === "length") {
            continue;
        }
        if (key === _propsTs.mappedKeys.index) {
            arrayProps.push(_numericLiteralsTs.wellFormedNonNegativeIntegerMatcher);
        } else {
            arrayProps.push(key);
        }
    }
    return arrayProps;
};
