"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    writeMalformedDistributableFunctionMessage: ()=>writeMalformedDistributableFunctionMessage,
    distributeFunctionToNode: ()=>distributeFunctionToNode
});
const _domainsTs = require("../../utils/domains.js");
const _errorsTs = require("../../utils/errors.js");
const _genericsTs = require("../../utils/generics.js");
const _serializeTs = require("../../utils/serialize.js");
const writeMalformedDistributableFunctionMessage = (def)=>`Expected a Function or Record<Domain, Function> operand (${(0, _serializeTs.stringify)(def)} was invalid)`;
const distributeFunctionToNode = (distributableFunction, inputNode, ctx, ruleKey)=>{
    const domains = (0, _genericsTs.keysOf)(ctx.type.meta.scope.resolveNode(inputNode));
    if (!(0, _domainsTs.hasDomain)(distributableFunction, "object")) {
        return (0, _errorsTs.throwParseError)(writeMalformedDistributableFunctionMessage(distributableFunction));
    }
    const distributed = {};
    if (typeof distributableFunction === "function") {
        const domainFunction = {
            [ruleKey]: distributableFunction
        };
        for (const domain of domains){
            distributed[domain] = domainFunction;
        }
    } else {
        for (const domain1 of domains){
            if (distributableFunction[domain1] === undefined) {
                continue;
            }
            const functionInDomain = {
                [ruleKey]: distributableFunction[domain1]
            };
            if (typeof functionInDomain[ruleKey] !== "function") {
                return (0, _errorsTs.throwParseError)(writeMalformedDistributableFunctionMessage(functionInDomain));
            }
            distributed[domain1] = functionInDomain;
        }
    }
    return distributed;
};
