"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    Scope: ()=>Scope,
    writeShallowCycleErrorMessage: ()=>writeShallowCycleErrorMessage,
    scope: ()=>scope,
    writeDuplicateAliasesMessage: ()=>writeDuplicateAliasesMessage
});
const _nodeTs = require("../nodes/node.js");
const _definitionTs = require("../parse/definition.js");
const _chainableNoOpProxyTs = require("../utils/chainableNoOpProxy.js");
const _errorsTs = require("../utils/errors.js");
const _pathsTs = require("../utils/paths.js");
const _cacheTs = require("./cache.js");
const _typeTs = require("./type.js");
function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classApplyDescriptorUpdate(receiver, descriptor) {
    if (descriptor.set) {
        if (!descriptor.get) {
            throw new TypeError("attempted to read set only private field");
        }
        if (!("__destrWrapper" in descriptor)) {
            descriptor.__destrWrapper = {
                set value (v){
                    descriptor.set.call(receiver, v);
                },
                get value () {
                    return descriptor.get.call(receiver);
                }
            };
        }
        return descriptor.__destrWrapper;
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        return descriptor;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
function _classPrivateFieldUpdate(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "update");
    return _classApplyDescriptorUpdate(receiver, descriptor);
}
function _classPrivateMethodGet(receiver, privateSet, fn) {
    if (!privateSet.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return fn;
}
function _classPrivateMethodInit(obj, privateSet) {
    _checkPrivateRedeclaration(obj, privateSet);
    privateSet.add(obj);
}
var _ctx_type_meta;
let anonymousScopeCount = 0;
const scopeRegistry = {};
const spaceRegistry = {};
var _resolutions = /*#__PURE__*/ new WeakMap(), _exports = /*#__PURE__*/ new WeakMap(), _anonymousTypeCount = /*#__PURE__*/ new WeakMap(), _register = /*#__PURE__*/ new WeakSet(), _cacheSpaces = /*#__PURE__*/ new WeakSet(), _initializeContext = /*#__PURE__*/ new WeakSet(), _resolveRecurse = /*#__PURE__*/ new WeakSet();
class Scope {
    createAnonymousTypeSuffix() {
        return ++_classPrivateFieldUpdate(this, _anonymousTypeCount).value;
    }
    get infer() {
        return _chainableNoOpProxyTs.chainableNoOpProxy;
    }
    compile() {
        if (!spaceRegistry[this.name]) {
            for(const name in this.aliases){
                this.resolve(name);
            }
            spaceRegistry[this.name] = _classPrivateFieldGet(this, _exports).root;
        }
        return _classPrivateFieldGet(this, _exports).root;
    }
    addReferenceIfResolvable(name, ctx) {
        const resolution = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "undefined", []);
        if (!resolution) {
            return false;
        }
        (_ctx_type_meta = ctx.type.meta).includesMorph || (_ctx_type_meta.includesMorph = resolution.meta.includesMorph);
        return true;
    }
    resolve(name) {
        return _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "throw", []);
    }
    resolveNode(node) {
        if (typeof node === "object") {
            return node;
        }
        return this.resolveNode(this.resolve(node).node);
    }
    constructor(aliases, opts){
        _classPrivateMethodInit(this, _register);
        _classPrivateMethodInit(this, _cacheSpaces);
        _classPrivateMethodInit(this, _initializeContext);
        _classPrivateMethodInit(this, _resolveRecurse);
        _classPrivateFieldInit(this, _resolutions, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _exports, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _anonymousTypeCount, {
            writable: true,
            value: void 0
        });
        this.aliases = aliases;
        this.opts = opts;
        this.parseCache = new _cacheTs.FreezingCache();
        _classPrivateFieldSet(this, _resolutions, new _cacheTs.Cache());
        _classPrivateFieldSet(this, _exports, new _cacheTs.Cache());
        _classPrivateFieldSet(this, _anonymousTypeCount, 0);
        this.type = (def, opts = {})=>{
            if (opts.name && this.aliases[opts.name]) {
                return (0, _errorsTs.throwParseError)(writeDuplicateAliasesMessage(opts.name));
            }
            const result = (0, _typeTs.initializeType)(def, opts, this);
            const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, result);
            result.node = this.resolveNode((0, _definitionTs.parseDefinition)(def, ctx));
            result.flat = (0, _nodeTs.flattenType)(result);
            return result;
        };
        this.name = _classPrivateMethodGet(this, _register, register).call(this, opts);
        if (opts.standard !== false) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, [
                spaceRegistry["standard"]
            ], "imports");
        }
        if (opts.imports) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.imports, "imports");
        }
        if (opts.includes) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.includes, "includes");
        }
        if (opts.config) {
            this.config = opts.config;
        }
    }
}
function register(opts) {
    const name = opts.name ? scopeRegistry[opts.name] ? (0, _errorsTs.throwParseError)(`A scope named '${opts.name}' already exists`) : opts.name : `scope${++anonymousScopeCount}`;
    scopeRegistry[name] = this;
    return name;
}
function cacheSpaces(spaces, kind) {
    for (const space of spaces){
        for(const name in space){
            if (_classPrivateFieldGet(this, _resolutions).has(name) || name in this.aliases) {
                (0, _errorsTs.throwParseError)(writeDuplicateAliasesMessage(name));
            }
            _classPrivateFieldGet(this, _resolutions).set(name, space[name]);
            if (kind === "includes") {
                _classPrivateFieldGet(this, _exports).set(name, space[name]);
            }
        }
    }
}
function initializeContext(type) {
    return {
        type,
        path: new _pathsTs.Path()
    };
}
function resolveRecurse(name, onUnresolvable, seen) {
    const maybeCacheResult = _classPrivateFieldGet(this, _resolutions).get(name);
    if (maybeCacheResult) {
        return maybeCacheResult;
    }
    if (!this.aliases[name]) {
        return onUnresolvable === "throw" ? (0, _errorsTs.throwInternalError)(`Unexpectedly failed to resolve alias '${name}'`) : undefined;
    }
    const type = (0, _typeTs.initializeType)(this.aliases[name], {
        name
    }, this);
    _classPrivateFieldGet(this, _resolutions).set(name, type);
    _classPrivateFieldGet(this, _exports).set(name, type);
    const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, type);
    let resolution = (0, _definitionTs.parseDefinition)(type.meta.definition, ctx);
    if (typeof resolution === "string") {
        if (seen.includes(resolution)) {
            return (0, _errorsTs.throwParseError)(writeShallowCycleErrorMessage(name, seen));
        }
        seen.push(resolution);
        resolution = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, resolution, "throw", seen).node;
    }
    type.node = resolution;
    type.flat = (0, _nodeTs.flattenType)(type);
    return type;
}
const writeShallowCycleErrorMessage = (name, seen)=>`Alias '${name}' has a shallow resolution cycle: ${[
        ...seen,
        name
    ].join("=>")}`;
const scope = (aliases, opts = {})=>new Scope(aliases, opts);
const writeDuplicateAliasesMessage = (name)=>`Alias '${name}' is already defined`;
