"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    initializeType: ()=>initializeType,
    isType: ()=>isType
});
const _traverseTs = require("../traverse/traverse.js");
const _chainableNoOpProxyTs = require("../utils/chainableNoOpProxy.js");
const _genericsTs = require("../utils/generics.js");
const compileTypeConfig = (opts)=>{
    if (opts === undefined) {
        return;
    }
    const { name , ...config } = opts;
    if ((0, _genericsTs.hasKeys)(config)) {
        return config;
    }
};
const initializeType = (definition, opts, scope)=>{
    const name = opts?.name ?? "type";
    const config = compileTypeConfig(opts);
    const meta = {
        name,
        id: `${scope.name}.${opts?.name ? name : `type${scope.createAnonymousTypeSuffix()}`}`,
        definition,
        scope,
        config,
        includesMorph: false
    };
    const root = {
        // temporarily initialize node/flat to aliases that will be included in
        // the final type in case of cyclic resolutions
        node: name,
        flat: [
            [
                "alias",
                name
            ]
        ],
        meta,
        infer: _chainableNoOpProxyTs.chainableNoOpProxy
    };
    // dynamically assign a name to the primary traversal function
    const namedTraverse = {
        [name]: (data)=>{
            const state = new _traverseTs.TraversalState(data, type);
            return (0, _traverseTs.traverse)(type.flat, state) ? {
                data: state.data
            } : {
                problems: state.problems
            };
        }
    }[name];
    // we need to assign this to a variable before returning so we can reference
    // it in namedTraverse
    const type = Object.assign(namedTraverse, root);
    return type;
};
const isType = (value)=>value?.infer === _chainableNoOpProxyTs.chainableNoOpProxy;
