"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    TraversalState: ()=>TraversalState,
    traverse: ()=>traverse,
    checkEntries: ()=>checkEntries,
    checkRequiredProp: ()=>checkRequiredProp
});
const _discriminateJs = require("../nodes/discriminate.js");
const _classJs = require("../nodes/rules/class.js");
const _divisorJs = require("../nodes/rules/divisor.js");
const _rangeJs = require("../nodes/rules/range.js");
const _regexJs = require("../nodes/rules/regex.js");
const _rulesJs = require("../nodes/rules/rules.js");
const _domainsJs = require("../utils/domains.js");
const _genericsJs = require("../utils/generics.js");
const _pathsJs = require("../utils/paths.js");
const _serializeJs = require("../utils/serialize.js");
const _problemsJs = require("./problems.js");
function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
var _seen = /*#__PURE__*/ new WeakMap();
class TraversalState {
    getConfigForProblemCode(code) {
        if (!this.configs.length) {
            return _problemsJs.defaultProblemWriters[code];
        }
        for(let i = this.configs.length - 1; i >= 0; i--){
            if (this.configs[i][code] || this.configs[i]["defaults"]) {
                return {
                    ..._problemsJs.defaultProblemWriters[code],
                    ...this.configs[i]["defaults"],
                    ...this.configs[i][code]
                };
            }
        }
        return _problemsJs.defaultProblemWriters[code];
    }
    traverseKey(key, node) {
        const lastData = this.data;
        this.data = this.data[key];
        // TODO: make path externally readonly
        this.path.push(key);
        const isValid = traverse(node, this);
        this.path.pop();
        lastData[key] = this.data;
        this.data = lastData;
        return isValid;
    }
    traverseResolution(name) {
        const resolution = this.type.meta.scope.resolve(name);
        const id = resolution.meta.id;
        // this assignment helps with narrowing
        const data = this.data;
        const isObject = (0, _domainsJs.hasDomain)(data, "object");
        if (isObject) {
            if ((0, _genericsJs.hasKey)(_classPrivateFieldGet(this, _seen), id)) {
                if (_classPrivateFieldGet(this, _seen)[id].includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return true;
                }
                _classPrivateFieldGet(this, _seen)[id].push(data);
            } else {
                _classPrivateFieldGet(this, _seen)[id] = [
                    data
                ];
            }
        }
        const lastResolution = this.type;
        this.type = resolution;
        const isValid = traverse(resolution.flat, this);
        this.type = lastResolution;
        if (isObject) {
            _classPrivateFieldGet(this, _seen)[id].pop();
        }
        return isValid;
    }
    traverseBranches(branches) {
        const lastFailFast = this.failFast;
        this.failFast = true;
        const lastProblems = this.problems;
        const branchProblems = new _problemsJs.Problems(this);
        this.problems = branchProblems;
        const lastPath = this.path;
        let hasValidBranch = false;
        for (const branch of branches){
            this.path = new _pathsJs.Path();
            if (checkEntries(branch, this)) {
                hasValidBranch = true;
                break;
            }
        }
        this.path = lastPath;
        this.problems = lastProblems;
        this.failFast = lastFailFast;
        return hasValidBranch || this.problems.add("branches", this.data, branchProblems);
    }
    constructor(data, type){
        _classPrivateFieldInit(this, _seen, {
            writable: true,
            value: void 0
        });
        this.data = data;
        this.type = type;
        this.path = new _pathsJs.Path();
        this.problems = new _problemsJs.Problems(this);
        this.failFast = false;
        _classPrivateFieldSet(this, _seen, {});
        this.configs = type.meta.scope.config ? [
            type.meta.scope.config
        ] : [];
    }
}
const traverse = (node, state)=>typeof node === "string" ? (0, _domainsJs.domainOf)(state.data) === node || state.problems.add("domain", state.data, node) : checkEntries(node, state);
const checkEntries = (entries, state)=>{
    let isValid = true;
    for(let i = 0; i < entries.length; i++){
        const [k, v] = entries[i];
        if (k === "morph") {
            if (typeof v === "function") {
                // TODO: allow problem from morph
                state.data = v(state.data);
            } else {
                for (const morph of v){
                    state.data = morph(state.data);
                }
            }
            return true;
        }
        const entryAllowsData = entryCheckers[k](v, state);
        isValid && (isValid = entryAllowsData);
        if (!isValid) {
            if (state.failFast) {
                return false;
            }
            if (i < entries.length - 1 && _rulesJs.precedenceMap[k] < _rulesJs.precedenceMap[entries[i + 1][0]]) {
                // if we've encountered a problem, there is at least one entry
                // remaining, and the next entry is of a higher precedence level
                // than the current entry, return immediately
                return false;
            }
        }
    }
    return isValid;
};
const checkRequiredProp = (prop, state)=>{
    if (prop[0] in state.data) {
        return state.traverseKey(prop[0], prop[1]);
    }
    return state.problems.add("missing", undefined, undefined, {
        path: state.path.concat(prop[0])
    });
};
const entryCheckers = {
    regex: _regexJs.checkRegex,
    divisor: _divisorJs.checkDivisor,
    domains: (domains, state)=>{
        const entries = domains[(0, _domainsJs.domainOf)(state.data)];
        return entries ? checkEntries(entries, state) : state.problems.add("domainBranches", state.data, (0, _genericsJs.keysOf)(domains));
    },
    // TODO: remove data from problem params
    domain: (domain, state)=>(0, _domainsJs.domainOf)(state.data) === domain || state.problems.add("domain", state.data, domain),
    bound: _rangeJs.checkBound,
    optionalProp: (prop, state)=>{
        if (prop[0] in state.data) {
            return state.traverseKey(prop[0], prop[1]);
        }
        return true;
    },
    // these checks work the same way, the keys are only distinct so that
    // prerequisite props can have a higher precedence
    requiredProp: checkRequiredProp,
    prerequisiteProp: checkRequiredProp,
    indexProp: (node, state)=>{
        if (!Array.isArray(state.data)) {
            return state.problems.add("class", state.data, "Array");
        }
        let isValid = true;
        for(let i = 0; i < state.data.length; i++){
            isValid && (isValid = state.traverseKey(`${i}`, node));
            if (!isValid && state.failFast) {
                return false;
            }
        }
        return isValid;
    },
    branches: (branches, state)=>state.traverseBranches(branches),
    switch: (rule, state)=>{
        const dataAtPath = (0, _pathsJs.getPath)(state.data, rule.path);
        const caseKey = (0, _discriminateJs.serializeCase)(rule.kind, dataAtPath);
        if ((0, _genericsJs.hasKey)(rule.cases, caseKey)) {
            return checkEntries(rule.cases[caseKey], state);
        }
        const caseKeys = (0, _genericsJs.keysOf)(rule.cases);
        const missingCasePath = state.path.concat(rule.path);
        return rule.kind === "value" ? state.problems.add("valueBranches", dataAtPath, caseKeys.map((k)=>(0, _serializeJs.deserializePrimitive)(k)), {
            path: missingCasePath
        }) : state.problems.add("domainBranches", dataAtPath, caseKeys, {
            path: missingCasePath
        });
    },
    alias: (name, state)=>state.traverseResolution(name),
    class: _classJs.checkClass,
    // TODO: fix
    narrow: (narrow, state)=>narrow(state.data, state.problems),
    config: ({ config , node  }, state)=>{
        state.configs.push(config);
        const result = traverse(node, state);
        state.configs.pop();
        return result;
    },
    value: (value, state)=>state.data === value || state.problems.add("value", state.data, value)
};
