"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    snapshot: ()=>snapshot,
    serialize: ()=>serialize,
    stringify: ()=>stringify,
    serializePrimitive: ()=>serializePrimitive,
    deserializePrimitive: ()=>deserializePrimitive
});
const _domainsTs = require("./domains.js");
const _genericsJs = require("./generics.js");
const _numericLiteralsJs = require("./numericLiterals.js");
const snapshot = (data, opts = {})=>snapshotRecurse(data, opts, []);
const snapshotRecurse = (v, context, seen)=>{
    switch((0, _domainsTs.domainOf)(v)){
        case "object":
            if (typeof v === "function") {
                return stringifyOpts.onFunction(v);
            }
            if (seen.includes(v)) {
                return "(cycle)";
            }
            const nextSeen = [
                ...seen,
                v
            ];
            if (Array.isArray(v)) {
                return v.map((item)=>snapshotRecurse(item, context, nextSeen));
            }
            const result = {};
            for(const k in v){
                result[k] = snapshotRecurse(v[k], context, nextSeen);
            }
            return result;
        case "symbol":
            return stringifyOpts.onSymbol(v);
        case "bigint":
            return `${v}n`;
        case "undefined":
            return "undefined";
        default:
            return v;
    }
};
const serialize = (data, opts = {}, indent)=>{
    const seen = [];
    return JSON.stringify(data, (k, v)=>{
        switch((0, _domainsTs.domainOf)(v)){
            case "object":
                if (seen.includes(v)) {
                    return opts.onCycle?.(v);
                }
                if (typeof v === "function") {
                    return opts.onFunction?.(v);
                }
                return v;
            case "undefined":
                return "undefined";
            case "symbol":
                return opts.onSymbol?.(v);
            case "bigint":
                return `${v}n`;
            default:
                return v;
        }
    }, indent);
};
const stringifyOpts = {
    onCycle: ()=>"(cycle)",
    onSymbol: (v)=>`(symbol${v.description && ` ${v.description}`})`,
    onFunction: (v)=>`(function${v.name && ` ${v.name}`})`
};
const stringify = (data, indent)=>{
    switch((0, _domainsTs.domainOf)(data)){
        case "object":
            return serialize(data, stringifyOpts, indent);
        case "symbol":
            return stringifyOpts.onSymbol(data);
        default:
            return serializePrimitive(data);
    }
};
const serializePrimitive = (value)=>typeof value === "string" ? `'${value}'` : typeof value === "bigint" ? `${value}n` : `${value}`;
const deserializePrimitive = (serialized)=>(0, _genericsJs.isKeyOf)(serialized, serializedKeywords) ? serializedKeywords[serialized] : serialized[0] === "'" ? serialized.slice(1, -1) : (0, _numericLiteralsJs.tryParseWellFormedBigint)(serialized) ?? (0, _numericLiteralsJs.tryParseWellFormedNumber)(serialized, true);
const serializedKeywords = {
    true: true,
    false: false,
    undefined,
    null: null
};
