import { isType } from "../scopes/type.js";
import { domainOf } from "../utils/domains.js";
import { throwParseError } from "../utils/errors.js";
import { objectKindOf } from "../utils/objectKinds.js";
import { stringify } from "../utils/serialize.js";
import { parseRecord } from "./record.js";
import { parseString } from "./string/string.js";
import { parseTuple } from "./tuple/tuple.js";
export const parseDefinition = (def, ctx)=>{
    const domain = domainOf(def);
    if (domain === "string") {
        return parseString(def, ctx);
    }
    if (domain !== "object") {
        return throwParseError(writeBadDefinitionTypeMessage(domain));
    }
    const objectKind = objectKindOf(def);
    switch(objectKind){
        case "Object":
            return parseRecord(def, ctx);
        case "Array":
            return parseTuple(def, ctx);
        case "RegExp":
            return {
                string: {
                    regex: def.source
                }
            };
        case "Function":
            return isType(def) ? def.node : throwParseError(writeBadDefinitionTypeMessage("Function"));
        default:
            return throwParseError(writeBadDefinitionTypeMessage(objectKind ?? stringify(def)));
    }
};
export const as = Symbol("as");
export const unknownDefinitionMessage = "Cannot statically parse a definition inferred as unknown. Consider using 'as inferred<...>' to cast it.";
export const writeBadDefinitionTypeMessage = (actual)=>`Type definitions must be strings or objects (was ${actual})`;
