import { resolutionExtendsDomain } from "../../nodes/predicate.js";
import { mappedKeys } from "../../nodes/rules/props.js";
import { throwParseError } from "../../utils/errors.js";
import { listFrom } from "../../utils/generics.js";
import { wellFormedNonNegativeIntegerMatcher } from "../../utils/numericLiterals.js";
import { parseDefinition } from "../definition.js";
import { writeImplicitNeverMessage } from "../string/ast.js";
const numericIndexKey = {
    regex: wellFormedNonNegativeIntegerMatcher.source
};
export const parseKeyOfTuple = (def, ctx)=>{
    const resolution = ctx.type.meta.scope.resolveNode(parseDefinition(def[1], ctx));
    if (!resolutionExtendsDomain(resolution, "object")) {
        return throwParseError(writeImplicitNeverMessage(ctx.path, "keyof"));
    }
    if (resolution.object === true) {
        return writeImplicitNeverMessage(ctx.path, "keyof");
    }
    const objectBranches = listFrom(resolution.object);
    let sharedKeys = keysOfBranch(objectBranches[0]);
    for(let i = 1; i < objectBranches.length; i++){
        const branchKeys = keysOfBranch(objectBranches[i]);
        // we can filter directly by equality here because the RegExp we're
        // using will always be reference equal to
        // wellFormedNonNegativeIntegerMatcher
        sharedKeys = sharedKeys.filter((k)=>branchKeys.includes(k));
    }
    const result = sharedKeys.map((k)=>typeof k === "string" ? {
            value: k
        } : {
            regex: k.source
        });
    return result.length ? {
        string: result.length === 1 ? result[0] : result
    } : writeImplicitNeverMessage(ctx.path, "keyof");
};
const keysOfBranch = (branch)=>{
    if (!("props" in branch)) {
        return [];
    }
    if (branch.class !== "Array") {
        return Object.keys(branch.props);
    }
    const arrayProps = [];
    for (const key of Object.keys(branch.props)){
        if (key === "length") {
            continue;
        }
        if (key === mappedKeys.index) {
            arrayProps.push(wellFormedNonNegativeIntegerMatcher);
        } else {
            arrayProps.push(key);
        }
    }
    return arrayProps;
};
