import { TraversalState, traverse } from "../traverse/traverse.js";
import { chainableNoOpProxy } from "../utils/chainableNoOpProxy.js";
import { hasKeys } from "../utils/generics.js";
const compileTypeConfig = (opts)=>{
    if (opts === undefined) {
        return;
    }
    const { name , ...config } = opts;
    if (hasKeys(config)) {
        return config;
    }
};
export const initializeType = (definition, opts, scope)=>{
    const name = opts?.name ?? "type";
    const config = compileTypeConfig(opts);
    const meta = {
        name,
        id: `${scope.name}.${opts?.name ? name : `type${scope.createAnonymousTypeSuffix()}`}`,
        definition,
        scope,
        config,
        includesMorph: false
    };
    const root = {
        // temporarily initialize node/flat to aliases that will be included in
        // the final type in case of cyclic resolutions
        node: name,
        flat: [
            [
                "alias",
                name
            ]
        ],
        meta,
        infer: chainableNoOpProxy
    };
    // dynamically assign a name to the primary traversal function
    const namedTraverse = {
        [name]: (data)=>{
            const state = new TraversalState(data, type);
            return traverse(type.flat, state) ? {
                data: state.data
            } : {
                problems: state.problems
            };
        }
    }[name];
    // we need to assign this to a variable before returning so we can reference
    // it in namedTraverse
    const type = Object.assign(namedTraverse, root);
    return type;
};
export const isType = (value)=>value?.infer === chainableNoOpProxy;
