import { domainOf } from "./domains.js";
import { isKeyOf } from "./generics.js";
import { tryParseWellFormedBigint, tryParseWellFormedNumber } from "./numericLiterals.js";
export const snapshot = (data, opts = {})=>snapshotRecurse(data, opts, []);
const snapshotRecurse = (v, context, seen)=>{
    switch(domainOf(v)){
        case "object":
            if (typeof v === "function") {
                return stringifyOpts.onFunction(v);
            }
            if (seen.includes(v)) {
                return "(cycle)";
            }
            const nextSeen = [
                ...seen,
                v
            ];
            if (Array.isArray(v)) {
                return v.map((item)=>snapshotRecurse(item, context, nextSeen));
            }
            const result = {};
            for(const k in v){
                result[k] = snapshotRecurse(v[k], context, nextSeen);
            }
            return result;
        case "symbol":
            return stringifyOpts.onSymbol(v);
        case "bigint":
            return `${v}n`;
        case "undefined":
            return "undefined";
        default:
            return v;
    }
};
export const serialize = (data, opts = {}, indent)=>{
    const seen = [];
    return JSON.stringify(data, (k, v)=>{
        switch(domainOf(v)){
            case "object":
                if (seen.includes(v)) {
                    return opts.onCycle?.(v);
                }
                if (typeof v === "function") {
                    return opts.onFunction?.(v);
                }
                return v;
            case "undefined":
                return "undefined";
            case "symbol":
                return opts.onSymbol?.(v);
            case "bigint":
                return `${v}n`;
            default:
                return v;
        }
    }, indent);
};
const stringifyOpts = {
    onCycle: ()=>"(cycle)",
    onSymbol: (v)=>`(symbol${v.description && ` ${v.description}`})`,
    onFunction: (v)=>`(function${v.name && ` ${v.name}`})`
};
export const stringify = (data, indent)=>{
    switch(domainOf(data)){
        case "object":
            return serialize(data, stringifyOpts, indent);
        case "symbol":
            return stringifyOpts.onSymbol(data);
        default:
            return serializePrimitive(data);
    }
};
export const serializePrimitive = (value)=>typeof value === "string" ? `'${value}'` : typeof value === "bigint" ? `${value}n` : `${value}`;
export const deserializePrimitive = (serialized)=>isKeyOf(serialized, serializedKeywords) ? serializedKeywords[serialized] : serialized[0] === "'" ? serialized.slice(1, -1) : tryParseWellFormedBigint(serialized) ?? tryParseWellFormedNumber(serialized, true);
const serializedKeywords = {
    true: true,
    false: false,
    undefined,
    null: null
};
