import type { Domain } from "../utils/domains.js";
import type { evaluate } from "../utils/generics.js";
import type { NumberLiteral } from "../utils/numericLiterals.js";
import type { DefaultObjectKind } from "../utils/objectKinds.js";
import { Path } from "../utils/paths.js";
import type { SerializedPrimitive } from "../utils/serialize.js";
import type { Branches } from "./branch.js";
import type { FlattenContext, TraversalEntry } from "./node.js";
export type DiscriminatedSwitch<kind extends DiscriminantKind = DiscriminantKind> = {
    readonly path: Path;
    readonly kind: kind;
    readonly cases: DiscriminatedCases<kind>;
};
export type DiscriminatedCases<kind extends DiscriminantKind = DiscriminantKind> = {
    [caseKey in CaseKey<kind>]?: TraversalEntry[];
};
export declare const flattenBranches: (branches: Branches, ctx: FlattenContext) => TraversalEntry[];
export type QualifiedDisjoint = `${DiscriminantKind}` | `${string}/${DiscriminantKind}`;
export type DiscriminantKinds = {
    domain: Domain;
    objectKind: DefaultObjectKind;
    value: unknown;
};
export type DiscriminantKind = evaluate<keyof DiscriminantKinds>;
export declare const serializeIfAllowed: <kind extends keyof DiscriminantKinds>(kind: kind, data: DiscriminantKinds[kind]) => CaseKey<kind> | undefined;
export declare const serializeCase: <kind extends keyof DiscriminantKinds>(kind: kind, data: unknown) => CaseKey<kind>;
type CaseKey<kind extends DiscriminantKind = DiscriminantKind> = kind extends "value" ? SerializedPrimitive | "default" : kind extends "tupleLength" ? NumberLiteral | "default" : DiscriminantKinds[kind];
export {};
