import type { ParseContext } from "../parse/definition.js";
import type { Type, TypeConfig } from "../scopes/type.js";
import type { Domain, inferDomain } from "../utils/domains.js";
import type { Dict, stringKeyOf } from "../utils/generics.js";
import type { MorphEntry } from "./branch.js";
import type { Intersector } from "./compose.js";
import type { DiscriminatedSwitch } from "./discriminate.js";
import type { Predicate } from "./predicate.js";
import type { LiteralRules, RuleEntry } from "./rules/rules.js";
export type TypeNode<$ = Dict> = Identifier<$> | ResolvedNode<$>;
export type Identifier<$ = Dict> = stringKeyOf<$>;
export type ResolvedNode<$ = Dict> = {
    readonly [domain in Domain]?: Predicate<domain, $>;
};
export declare const nodeIntersection: Intersector<TypeNode>;
export declare const intersection: (l: TypeNode, r: TypeNode, type: Type) => TypeNode;
export declare const union: (l: TypeNode, r: TypeNode, type: Type) => ResolvedNode;
export type TraversalNode = Domain | TraversalEntry[];
export type TraversalKey = TraversalEntry[0];
export type TraversalValue<k extends TraversalKey> = Extract<TraversalEntry, [
    k,
    unknown
]>[1];
export type TraversalEntry = RuleEntry | DomainsEntry | MorphEntry | CyclicReferenceEntry | DomainEntry | BranchesEntry | SwitchEntry | ConfigEntry;
export type CyclicReferenceEntry = ["alias", string];
export type ConfigEntry = [
    "config",
    {
        config: TypeConfig;
        node: TraversalNode;
    }
];
export type DomainEntry = ["domain", Domain];
export type DomainsEntry = [
    "domains",
    {
        readonly [domain in Domain]?: TraversalEntry[];
    }
];
export type BranchesEntry = ["branches", TraversalEntry[][]];
export type SwitchEntry = ["switch", DiscriminatedSwitch];
export type FlattenContext = ParseContext & {
    lastDomain: Domain;
};
export declare const flattenType: (type: Type) => TraversalNode;
export declare const flattenNode: (node: TypeNode, ctx: FlattenContext) => TraversalNode;
export type LiteralNode<domain extends Domain = Domain, value extends inferDomain<domain> = inferDomain<domain>> = {
    [k in domain]: LiteralRules<domain, value>;
};
export declare const isLiteralNode: <domain extends keyof {
    bigint: bigint;
    boolean: boolean;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    undefined: undefined;
    null: null;
}>(node: ResolvedNode, domain: domain) => node is LiteralNode<domain, inferDomain<domain>>;
export declare const arrayOf: (node: TypeNode) => ResolvedNode;
