import type { Dict } from "../../utils/generics.js";
import type { TraversalNode, TypeNode } from "../node.js";
import type { FlattenAndPushRule } from "./rules.js";
export type PropsRule<$ = Dict> = {
    [propKey in string]: Prop<$>;
};
export type Prop<$ = Dict, node extends TypeNode<$> = TypeNode<$>> = node | OptionalProp<$, node> | PrerequisiteProp<$, node>;
export type OptionalProp<$ = Dict, node extends TypeNode<$> = TypeNode<$>> = [
    "?",
    node
];
export type PrerequisiteProp<$ = Dict, node extends TypeNode<$> = TypeNode<$>> = ["!", node];
export type PropEntry = RequiredPropEntry | OptionalPropEntry | IndexPropEntry | PrerequisitePropEntry;
export type PrerequisitePropEntry = ["prerequisiteProp", TraversalProp];
export type RequiredPropEntry = ["requiredProp", TraversalProp];
export type OptionalPropEntry = ["optionalProp", TraversalProp];
export type IndexPropEntry = ["indexProp", TraversalNode];
export type TraversalProp<key extends string = string, node extends TraversalNode = TraversalNode> = [key, node];
export declare const isOptional: (prop: Prop) => prop is OptionalProp<Dict, TypeNode<Dict>>;
export declare const isPrerequisite: (prop: Prop) => prop is PrerequisiteProp<Dict, TypeNode<Dict>>;
export declare const mappedKeys: {
    readonly index: "[index]";
};
export type MappedKeys = typeof mappedKeys;
export type MappedPropKey = MappedKeys[keyof MappedKeys];
export declare const propToNode: (prop: Prop) => TypeNode<Dict>;
export declare const propsIntersection: import("../compose.js").Intersector<PropsRule<Dict> | undefined>;
export declare const flattenProps: FlattenAndPushRule<PropsRule>;
