import type { TypeNode } from "../nodes/node.js";
import type { Type } from "../scopes/type.js";
import type { Primitive } from "../utils/domains.js";
import { domainOf } from "../utils/domains.js";
import type { Dict, evaluateObject, isAny, isUnknown, List } from "../utils/generics.js";
import { objectKindOf } from "../utils/objectKinds.js";
import type { Path } from "../utils/paths.js";
import type { inferRecord } from "./record.js";
import type { inferString, validateString } from "./string/string.js";
import type { inferTuple, TupleExpression, validateTupleExpression } from "./tuple/tuple.js";
export type ParseContext = {
    type: Type;
    path: Path;
};
export declare const parseDefinition: (def: unknown, ctx: ParseContext) => TypeNode;
export type inferDefinition<def, $> = isAny<def> extends true ? never : def extends inferred<infer t> ? t : def extends string ? inferString<def, $> : def extends List ? inferTuple<def, $> : def extends RegExp ? string : def extends Dict ? inferRecord<def, $> : never;
export type validateDefinition<def, $> = def extends Terminal ? def : def extends string ? validateString<def, $> : def extends TupleExpression ? validateTupleExpression<def, $> : def extends BadDefinitionType ? writeBadDefinitionTypeMessage<objectKindOf<def> extends string ? objectKindOf<def> : domainOf<def>> : isUnknown<def> extends true ? unknownDefinitionMessage : evaluateObject<{
    [k in keyof def]: validateDefinition<def[k], $>;
}>;
export declare const as: unique symbol;
export type inferred<t> = {
    [as]?: t;
};
export declare const unknownDefinitionMessage = "Cannot statically parse a definition inferred as unknown. Consider using 'as inferred<...>' to cast it.";
export type unknownDefinitionMessage = typeof unknownDefinitionMessage;
type Terminal = RegExp | inferred<unknown>;
type BadDefinitionType = Exclude<Primitive, string> | Function;
export declare const writeBadDefinitionTypeMessage: <actual extends string>(actual: actual) => `Type definitions must be strings or objects (was ${actual})`;
type writeBadDefinitionTypeMessage<actual extends string> = `Type definitions must be strings or objects (was ${actual})`;
export {};
