import type { DisjointsByPath } from "../../nodes/compose.js";
import type { MappedKeys } from "../../nodes/rules/props.js";
import type { resolve } from "../../scopes/scope.js";
import type { asIn } from "../../scopes/type.js";
import type { domainOf } from "../../utils/domains.js";
import type { asConst, castOnError, Dict, equals, error, evaluate, evaluateObject, extractValues, isAny, List, Literalable, RegexLiteral, requiredKeyOf, stringKeyOf, tryCatch } from "../../utils/generics.js";
import type { objectKindOf } from "../../utils/objectKinds.js";
import type { Path, pathToString } from "../../utils/paths.js";
import type { SizedData } from "../../utils/size.js";
import type { Out, ParsedMorph } from "../tuple/morph.js";
import type { StringLiteral } from "./shift/operand/enclosed.js";
import type { Scanner } from "./shift/scanner.js";
export type inferAst<ast, $> = ast extends List ? ast[1] extends "[]" ? inferAst<ast[0], $>[] : ast[1] extends "|" ? inferUnion<inferAst<ast[0], $>, inferAst<ast[2], $>> extends infer result ? castOnError<result, never> : never : ast[1] extends "&" ? inferIntersection<inferAst<ast[0], $>, inferAst<ast[2], $>> extends infer result ? castOnError<result, never> : never : ast[1] extends Scanner.Comparator ? ast[0] extends number ? inferAst<ast[2], $> : inferAst<ast[0], $> : ast[1] extends "%" ? inferAst<ast[0], $> : never : inferTerminal<ast, $>;
export type validateAstSemantics<ast, $> = ast extends string ? undefined : ast extends [infer child, unknown] ? validateAstSemantics<child, $> : ast extends [infer l, infer token, infer r] ? token extends "&" ? tryCatch<inferIntersection<inferAst<l, $>, inferAst<r, $>>, validateBinary<l, r, $>> : token extends "|" ? tryCatch<inferUnion<inferAst<l, $>, inferAst<r, $>>, validateBinary<l, r, $>> : token extends Scanner.Comparator ? l extends number ? validateAstSemantics<r, $> : isBoundable<inferAst<l, $>> extends true ? validateAstSemantics<l, $> : error<writeUnboundableMessage<astToString<ast[0]>>> : token extends "%" ? isDivisible<inferAst<l, $>> extends true ? validateAstSemantics<l, $> : error<writeIndivisibleMessage<astToString<ast[0]>>> : validateAstSemantics<l, $> : undefined;
type validateBinary<l, r, $> = tryCatch<validateAstSemantics<l, $>, tryCatch<validateAstSemantics<r, $>, undefined>>;
export type inferIntersection<l, r> = inferIntersectionRecurse<l, r, []>;
type inferIntersectionRecurse<l, r, path extends string[]> = path["length"] extends 10 ? l & r : l extends never ? never : r extends never ? never : l & r extends never ? error<writeImplicitNeverMessage<path, "Intersection">> : isAny<l | r> extends true ? any : l extends ParsedMorph<infer lIn, infer lOut> ? r extends ParsedMorph ? error<writeImplicitNeverMessage<path, "Intersection", "of morphs">> : (In: evaluate<lIn & r>) => Out<lOut> : r extends ParsedMorph<infer rIn, infer rOut> ? (In: evaluate<rIn & l>) => Out<rOut> : [l, r] extends [Dict, Dict] ? bubblePropErrors<evaluateObject<{
    [k in stringKeyOf<l>]: k extends keyof r ? inferIntersectionRecurse<l[k], r[k], [...path, k]> : l[k];
} & Omit<r, keyof l>>> : [l, r] extends [List<infer lItem>, List<infer rItem>] ? inferIntersectionRecurse<lItem, rItem, [
    ...path,
    MappedKeys["index"]
]> extends infer result ? tryCatch<result, result[]> : never : l & r;
type bubblePropErrors<o> = extractValues<o, error> extends never ? o : extractValues<o, error>;
export type inferUnion<l, r> = isAny<l | r> extends true ? any : [l] extends [never] ? r : [r] extends [never] ? l : [asIn<l>, asIn<r>] extends [infer lIn, infer rIn] ? [equals<l, lIn>, equals<r, rIn>] extends [true, true] ? l | r : discriminatable<lIn, rIn> extends true ? l | r : error<undiscriminatableMorphUnionMessage> : never;
type discriminatable<l, r> = discriminatableRecurse<l, r, []> extends never ? false : true;
type discriminatableRecurse<l, r, path extends string[]> = path["length"] extends 10 ? never : l & r extends never ? path : domainOf<l> & domainOf<r> extends never ? path : objectKindOf<l> & objectKindOf<r> extends never ? path : [objectKindOf<l>, objectKindOf<r>] extends ["Object", "Object"] ? extractValues<{
    [k in requiredKeyOf<l>]: k extends requiredKeyOf<r> ? discriminatableRecurse<l[k], r[k], [...path, k & string]> : never;
}, string[]> : never;
export declare const compileDisjointReasonsMessage: (disjoints: DisjointsByPath) => string;
export declare const writeImplicitNeverMessage: <path extends [] | Path, operator extends "keyof" | "Intersection", description extends string = "">(path: asConst<path>, operator: operator, description?: description | undefined) => `${path extends [] ? "" : `At ${pathToString<path, "/">}: `}${operator} ${description extends "" ? "" : `${description} `}results in an unsatisfiable type`;
export type writeImplicitNeverMessage<path extends string[], operator extends "Intersection" | "keyof", description extends string = ""> = `${path extends [] ? "" : `At ${pathToString<path>}: `}${operator} ${description extends "" ? "" : `${description} `}results in an unsatisfiable type`;
export declare const writeUndiscriminatableMorphUnionMessage: <path extends string>(path: path) => string;
type undiscriminatableMorphUnionMessage = `A union including one or more morphs must be discriminatable`;
type isDivisible<data> = isAny<data> extends true ? false : [data] extends [number] ? true : false;
type isBoundable<data> = isAny<data> extends true ? false : [data] extends [SizedData] ? true : false;
export type inferTerminal<token, $> = token extends keyof $ ? resolve<token, $> : token extends StringLiteral<infer Text> ? Text : token extends RegexLiteral ? string : token extends number | bigint ? token : never;
export type astToString<ast> = `'${astToStringRecurse<ast, "">}'`;
type astToStringRecurse<ast, result extends string> = ast extends [
    infer head,
    ...infer tail
] ? astToStringRecurse<tail, `${result}${astToStringRecurse<head, "">}`> : ast extends Literalable ? `${result}${ast extends bigint ? `${ast}n` : ast}` : "...";
export declare const writeIndivisibleMessage: <root extends string>(root: root) => `Divisibility operand ${root} must be a number`;
type writeIndivisibleMessage<root extends string> = `Divisibility operand ${root} must be a number`;
export declare const writeUnboundableMessage: <root extends string>(root: root) => `Bounded expression ${root} must be a number, string or array`;
type writeUnboundableMessage<root extends string> = `Bounded expression ${root} must be a number, string or array`;
export {};
