import type { inferNode } from "../../nodes/infer.js";
import type { ResolvedNode, TypeNode } from "../../nodes/node.js";
import type { asIn, asOut } from "../../scopes/type.js";
import type { conform, constructor, error, List, returnOf } from "../../utils/generics.js";
import type { inferDefinition, ParseContext, validateDefinition } from "../definition.js";
import type { inferIntersection, inferUnion } from "../string/ast.js";
import { writeMissingRightOperandMessage } from "../string/shift/operand/unenclosed.js";
import type { Scanner } from "../string/shift/scanner.js";
import type { inferKeyOfExpression, validateKeyOfExpression } from "./keyof.js";
import type { Out, validateMorphTuple } from "./morph.js";
import type { validateNarrowTuple } from "./narrow.js";
export declare const parseTuple: (def: List, ctx: ParseContext) => TypeNode;
export type validateTupleExpression<def extends TupleExpression, $> = def[1] extends "=>" ? validateMorphTuple<def, $> : def[1] extends ":" ? validateNarrowTuple<def, $> : def[1] extends Scanner.BranchToken ? def[2] extends undefined ? [def[0], error<writeMissingRightOperandMessage<def[1], "">>] : conform<def, readonly [
    validateDefinition<def[0], $>,
    def[1],
    validateDefinition<def[2], $>
]> : def[1] extends "[]" ? conform<def, readonly [validateDefinition<def[0], $>, "[]"]> : def[0] extends "===" ? conform<def, readonly ["===", unknown]> : def[0] extends "instanceof" ? conform<def, readonly ["instanceof", constructor]> : def[0] extends "node" ? conform<def, readonly ["node", ResolvedNode<$>]> : def[0] extends "keyof" ? validateKeyOfExpression<def[1], $> : never;
export type inferTuple<def extends List, $> = def extends TupleExpression ? inferTupleExpression<def, $> : {
    [i in keyof def]: inferDefinition<def[i], $>;
};
export type inferTupleExpression<def extends TupleExpression, $> = def[1] extends ":" ? "3" extends keyof def ? inferDefinition<def[3], $> : inferDefinition<def[0], $> : def[1] extends "=>" ? (In: asIn<inferDefinition<def[0], $>>) => Out<"3" extends keyof def ? asOut<inferDefinition<def[3], $>> : returnOf<def[2]>> : def[1] extends "&" ? inferIntersection<inferDefinition<def[0], $>, inferDefinition<def[2], $>> : def[1] extends "|" ? inferUnion<inferDefinition<def[0], $>, inferDefinition<def[2], $>> : def[1] extends "[]" ? inferDefinition<def[0], $>[] : def[0] extends "===" ? def[1] : def[0] extends "instanceof" ? def[1] extends constructor<infer t> ? t : never : def[0] extends "node" ? def[1] extends ResolvedNode<$> ? inferNode<def[1], $> : never : def[0] extends "keyof" ? inferKeyOfExpression<def[1], $> : never;
export type PostfixParser<token extends PostfixToken> = (def: PostfixExpression<token>, ctx: ParseContext) => TypeNode;
export type PrefixParser<token extends PrefixToken> = (def: PrefixExpression<token>, ctx: ParseContext) => TypeNode;
export type TupleExpression = PrefixExpression | PostfixExpression;
export declare const writeMalformedFunctionalExpressionMessage: (operator: "=>" | ":", rightDef: unknown) => string;
export type TupleExpressionToken = PrefixToken | PostfixToken;
type PostfixToken = "[]" | "&" | "|" | ":" | "=>";
export type PostfixExpression<token extends PostfixToken = PostfixToken> = readonly [unknown, token, ...unknown[]];
type PrefixToken = "keyof" | "instanceof" | "===" | "node";
export type PrefixExpression<token extends PrefixToken = PrefixToken> = readonly [token, ...unknown[]];
export {};
