import type { ResolvedNode, TypeNode } from "../nodes/node.js";
import type { inferDefinition, ParseContext, validateDefinition } from "../parse/definition.js";
import type { Dict, error, evaluateObject, isAny, List, nominal } from "../utils/generics.js";
import type { stringifyUnion } from "../utils/unionToTuple.js";
import { FreezingCache } from "./cache.js";
import type { PrecompiledDefaults } from "./standard.js";
import type { Type, TypeConfig, TypeParser } from "./type.js";
type ScopeParser = {
    <aliases>(aliases: validateAliases<aliases, {}>): Scope<parseScope<aliases, {}>>;
    <aliases, opts extends ScopeOptions>(aliases: validateAliases<aliases, opts>, opts: validateOptions<opts>): Scope<parseScope<aliases, opts>>;
};
export type ScopeOptions = {
    imports?: Space[] | [];
    includes?: Space[] | [];
    standard?: boolean;
    name?: string;
    config?: TypeConfig;
};
type validateOptions<opts extends ScopeOptions> = {
    [k in keyof opts]: k extends "imports" | "includes" ? mergeSpaces<opts[k], [
        k,
        "imports"
    ] extends ["includes", keyof opts] ? mergeSpaces<opts["includes"]> : {}> extends error<infer e> ? e : opts[k] : opts[k];
};
export type ScopeContext = Dict | ScopeContextTuple;
type ScopeContextTuple = [exports: Dict, locals: Dict, standard?: false];
type parseScope<aliases, opts extends ScopeOptions> = opts["standard"] extends false ? [inferExports<aliases, opts>, importsOf<opts>, false] : opts["imports"] extends Space[] ? [inferExports<aliases, opts>, importsOf<opts>] : inferExports<aliases, opts>;
type importsOf<opts extends ScopeOptions> = unknown extends opts["imports"] ? {} : mergeSpaces<opts["imports"]>;
type includesOf<opts extends ScopeOptions> = unknown extends opts["includes"] ? {} : mergeSpaces<opts["includes"]>;
export type resolve<name extends keyof $, $> = isAny<$[name]> extends true ? any : $[name] extends alias<infer def> ? inferDefinition<def, $> : $[name];
type exportsOf<context extends ScopeContext> = context extends [
    infer exports,
    ...unknown[]
] ? exports : context;
type localsOf<context extends ScopeContext> = context extends List ? context["1"] & (context["2"] extends false ? {} : PrecompiledDefaults) : PrecompiledDefaults;
type mergeSpaces<scopes, base extends Dict = {}> = scopes extends readonly [
    Space<infer head>,
    ...infer tail
] ? keyof base & keyof head extends never ? mergeSpaces<tail, base & head> : error<writeDuplicateAliasesMessage<stringifyUnion<keyof base & keyof head & string>>> : base;
type validateAliases<aliases, opts extends ScopeOptions> = {
    [name in keyof aliases]: name extends keyof preresolved<opts> ? writeDuplicateAliasesMessage<name & string> : validateDefinition<aliases[name], bootstrapScope<aliases, opts>>;
};
type preresolved<opts extends ScopeOptions> = includesOf<opts> & importsOf<opts> & (opts["standard"] extends false ? {} : PrecompiledDefaults);
type alias<def = {}> = nominal<def, "alias">;
type bootstrapScope<aliases, opts extends ScopeOptions> = {
    [k in keyof aliases]: alias<aliases[k]>;
} & preresolved<opts>;
type inferExports<aliases, opts extends ScopeOptions> = evaluateObject<{
    [k in keyof aliases]: inferDefinition<aliases[k], bootstrapScope<aliases, opts>>;
} & includesOf<opts>>;
export type Space<exports = Dict> = {
    [k in keyof exports]: Type<exports[k]>;
};
type resolutions<context extends ScopeContext> = localsOf<context> & exportsOf<context>;
type name<context extends ScopeContext> = keyof resolutions<context> & string;
export declare class Scope<context extends ScopeContext = any> {
    #private;
    aliases: Dict;
    opts: ScopeOptions;
    name: string;
    config: TypeConfig | undefined;
    parseCache: FreezingCache<TypeNode>;
    constructor(aliases: Dict, opts: ScopeOptions);
    type: TypeParser<resolutions<context>>;
    createAnonymousTypeSuffix(): number;
    get infer(): exportsOf<context>;
    compile(): Space<exportsOf<context>>;
    addReferenceIfResolvable(name: name<context>, ctx: ParseContext): boolean;
    resolve(name: name<context>): Type;
    resolveNode(node: TypeNode): ResolvedNode;
}
export declare const writeShallowCycleErrorMessage: (name: string, seen: string[]) => string;
export declare const scope: ScopeParser;
export declare const writeDuplicateAliasesMessage: <name_1 extends string>(name: name_1) => `Alias '${name_1}' is already defined`;
type writeDuplicateAliasesMessage<name extends string> = `Alias '${name}' is already defined`;
export {};
