import type { TraversalNode, TypeNode } from "../nodes/node.js";
import type { as, inferDefinition, validateDefinition } from "../parse/definition.js";
import type { ParsedMorph } from "../parse/tuple/morph.js";
import type { Problems, ProblemsConfig } from "../traverse/problems.js";
import type { defer, evaluateObject, xor } from "../utils/generics.js";
import type { BuiltinClass } from "../utils/objectKinds.js";
import type { Scope } from "./scope.js";
export type TypeParser<$> = {
    <def>(def: validateDefinition<def, $>): parseType<def, $>;
    <def>(def: validateDefinition<def, $>, opts: TypeOptions): parseType<def, $>;
};
export type parseType<def, $> = def extends validateDefinition<def, $> ? Type<inferDefinition<def, $>> : never;
type TypeRoot<t = unknown> = {
    [as]: t;
    infer: asOut<t>;
    node: TypeNode;
    flat: TraversalNode;
    meta: TypeMeta;
};
export type TypeOptions = evaluateObject<{
    name?: string;
} & ProblemsConfig>;
export type TypeConfig = ProblemsConfig;
type TypeMeta = {
    name: string;
    id: QualifiedTypeName;
    definition: unknown;
    scope: Scope;
    config: TypeConfig | undefined;
    includesMorph: boolean;
};
export declare const initializeType: (definition: unknown, opts: TypeOptions | undefined, scope: Scope) => Type<unknown>;
export declare const isType: (value: unknown) => value is Type<unknown>;
export type CheckResult<t> = xor<{
    data: asOut<t>;
}, {
    problems: Problems;
}>;
type Checker<t> = (data: unknown) => CheckResult<t>;
export type Type<t = unknown> = defer<Checker<t> & TypeRoot<t>>;
export type QualifiedTypeName = `${string}.${string}`;
export type asIn<t> = asIo<t, "in">;
export type asOut<t> = asIo<t, "out">;
type asIo<t, io extends "in" | "out"> = t extends ParsedMorph<infer i, infer o> ? io extends "in" ? i : o : t extends object ? t extends Function | BuiltinClass ? t : {
    [k in keyof t]: asIo<t[k], io>;
} : t;
export {};
