import type { TraversalEntry, TraversalKey, TraversalNode, TraversalValue } from "../nodes/node.js";
import type { TraversalProp } from "../nodes/rules/props.js";
import type { Type, TypeConfig } from "../scopes/type.js";
import type { extend, stringKeyOf } from "../utils/generics.js";
import { Path } from "../utils/paths.js";
import type { SizedData } from "../utils/size.js";
import type { ProblemCode, ProblemWriters } from "./problems.js";
export declare class TraversalState<data = unknown> {
    #private;
    data: data;
    type: Type;
    path: Path;
    problems: readonly import("./problems.js").Problem<any>[] & {
        byPath: Record<string, import("./problems.js").Problem<any>>;
        add: <code extends keyof {
            divisor: number;
            class: import("../utils/generics.js").constructor | "Array" | "Date" | "Error" | "Function" | "Map" | "RegExp" | "Set" | "Object" | "String" | "Number" | "Boolean" | "WeakMap" | "WeakSet" | "Promise";
            domain: keyof {
                bigint: bigint;
                boolean: boolean;
                number: number;
                object: object;
                string: string;
                symbol: symbol;
                undefined: undefined;
                null: null;
            };
            domainBranches: (keyof {
                bigint: bigint;
                boolean: boolean;
                number: number;
                object: object;
                string: string;
                symbol: symbol;
                undefined: undefined;
                null: null;
            })[];
            missing: undefined;
            bound: import("../nodes/rules/range.js").FlatBound;
            regex: `/${string}/`;
            value: unknown;
            valueBranches: unknown[];
            multi: import("./problems.js").Problem<any>[];
            branches: readonly import("./problems.js").Problem<any>[];
        }>(code: code, data: import("./problems.js").ProblemData<code>, source: import("./problems.js").ProblemSource<code>, opts?: import("./problems.js").AddProblemOptions | undefined) => false;
        summary: string;
        throw: () => never;
    };
    configs: TypeConfig[];
    failFast: boolean;
    constructor(data: data, type: Type);
    getConfigForProblemCode<code extends ProblemCode>(code: code): ProblemWriters<code>;
    traverseKey(key: stringKeyOf<this["data"]>, node: TraversalNode): boolean;
    traverseResolution(name: string): boolean;
    traverseBranches(branches: TraversalEntry[][]): boolean;
}
export declare const traverse: (node: TraversalNode, state: TraversalState) => boolean;
export declare const checkEntries: (entries: TraversalEntry[], state: TraversalState) => boolean;
export declare const checkRequiredProp: (prop: TraversalProp, state: TraversalState<TraversableData>) => boolean;
export type ValidationTraversalKey = Exclude<TraversalKey, "morph">;
export type EntryChecker<k extends TraversalKey> = (constraint: TraversalValue<k>, state: TraversalState<RuleData<k>>) => boolean;
export type TraversableData = Record<string | number, unknown>;
export type ConstrainedRuleTraversalData = extend<{
    [k in TraversalKey]?: unknown;
}, {
    regex: string;
    divisor: number;
    bound: SizedData;
    prerequisiteProp: TraversableData;
    optionalProp: TraversableData;
    requiredProp: TraversableData;
    indexProp: TraversableData;
}>;
export type RuleData<k extends TraversalKey> = k extends keyof ConstrainedRuleTraversalData ? ConstrainedRuleTraversalData[k] : unknown;
